require 'rails_helper'
module Hermes
  require [Hermes::Engine.root, %w[lib hermes tools specs]].flatten.join('/')

  MUTATE_PARAMS = {
    title: 'String',
    siret: 'String',
    siren: 'String',
    automated_request_validations: 'Boolean',
    delegated_validations: 'Boolean',
    priorities_enabled: 'Boolean',
    in_peec: 'Boolean',
    achieved_goal: 'Boolean',
    spent_on_completed_requests: 'Integer',
    spent_on_ongoing_requests: 'Integer',
    spent_on_other: 'Integer',
    annual_stock: 'Integer'
  }

  ACCEPTABLE_PARAMS = %w[
    als_id
    title
    siret
    siren
    is_root
    root_id
    automated_request_validations
    delegated_validations
    priorities_enabled
    in_peec
    achieved_goal
    spent_on_completed_requests
    spent_on_ongoing_requests
    spent_on_other
    annual_stock
    total_spent
    points_balance
    applicant_ids
    lessor_id
    group_id
    supervisor_group_id
    manager_group_id
  ]

  VALID_OPERATORS = %w[
    $in
    $nin
    $ne
    $gt
    $gte
    $lt
    $lte
    $eq
    $regex
    $all
  ]

  describe Api::Dmo::EntitiesController, :type => :api do
    routes {Hermes::Engine.routes}
    include Devise::Test::ControllerHelpers
    include JwtAuthHelper

    let!(:responsible) {create(:user, roles: ['Administrator'])}
    let!(:user) {create(:user)}

    describe 'POST /entities' do
      describe 'Create entity with minimal valid datas' do
        before :each do
          authenticated_header(request, responsible)
          post 'create', {
              data: {
                  attributes: {
                  }
              }
          }
        end
        it 'returns status code 201' do
          expect(response).to have_http_status(201)
        end
      end
      MUTATE_PARAMS.each do |param, type|
        describe 'post request with ' + param.to_s + ' (' + type.to_s + ') is valid' do
          let(:params) do
            attr = {}
            attr[param] = get_typed_value(type)
            attr
          end
          before :each do
            authenticated_header(request, responsible)
            post 'create', {
                data: {
                    attributes: params
                }
            }
          end
          it 'returns status code 201 and the value appears in json' do
            expect(response).to have_http_status(201)
            if (type == 'Date' || type == 'DateTime')
              expect(json['data']['attributes'][param.to_s]).not_to be_empty
              expect(json['data']['attributes'][param.to_s]).not_to eq('')
            else
              expect(json['data']['attributes'][param.to_s]).to eq(get_typed_value(type))
            end
          end
        end
      end
      #applicant_ids: 'Array',
      describe 'post request with applicant_ids' do
        let!(:applicant2) {create(:applicant)}
        let!(:applicant) {create(:applicant)}
        let!(:applicant2) {create(:applicant)}
        before :each do
          authenticated_header(request, responsible)
          post 'create', {
              data: {
                  attributes: {
                      applicant_ids: [applicant.id.to_s, applicant2.id.to_s]
                  }
              }
          }
        end
        it 'returns status code 201 and the value appears in json' do
          expect(response).to have_http_status(201)
          #TODO on pourrait s'attendre a retrouver les ids des housings requests mais c'est coherent car elles sont deja en belongs to avec un autre applicant
          expect(json['data']['attributes']['applicant_ids']).to eq([])
        end
      end
      # TODO: housing_offer_ids: 'Array',
      # describe 'post request with housing_offer_ids' do
      #   let!(:applicant2) {create(:applicant)}
      #   let!(:housing_offer) {create(:housing_offer)}
      #   let!(:housing_offer2) {create(:housing_offer)}
      #   before :each do
      #     authenticated_header(request, responsible)
      #     post 'create', {
      #         data: {
      #             attributes: {
      #                 housing_offer_ids: [housing_offer.id.to_s, housing_offer2.id.to_s]
      #             }
      #         }
      #     }
      #   end
      #   it 'returns status code 201 and the value appears in json' do
      #     expect(response).to have_http_status(201)
      #     #TODO on pourrait s'attendre a retrouver les ids des housings requests mais c'est coherent car elles sont deja en belongs to avec un autre applicant
      #     expect(json['data']['attributes']['housing_offer_ids']).to eq([])
      #   end
      # end
      #lessor_id: 'String'
      describe 'post request with lessor_id is valid' do
        let!(:lessor) {create(:lessor, user_id: responsible.id, title: 'Lessor')}
        before :each do
          authenticated_header(request, responsible)
          post 'create', {
              data: {
                  attributes: {lessor_id: lessor.id.to_s}
              }
          }
        end
        it 'returns status code 201 and the value appears in json' do
          expect(response).to have_http_status(201)
          expect(json['data']['attributes']['lessor_id']).to eq(lessor.id.to_s)
        end
      end
    end

    describe 'GET /entities' do
      describe 'GET /entities' do
        let!(:entity1) {create(:entity)}
        let!(:entity3) {create(:entity)}
        let!(:entity2) {create(:entity)}
        before :each do
          authenticated_header(request, responsible)
          get 'index'
        end
        it 'returns status code 200' do
          expect(response).to have_http_status(200)
        end
        it 'returns a valid json' do
          expect(json).not_to be_empty
          expect(json['data'].length).to eq(3)
          expect(json['meta']['pagination']).to be_present
          expect(json['meta']['pagination']['total_objects']).to eq(3)
        end

        describe 'GET /entities with batch of ids' do
          let!(:entity1) {create(:entity)}
          let!(:entity3) {create(:entity)}
          let!(:entity2) {create(:entity)}
          before :each do
            authenticated_header(request, responsible)
            get 'index', {ids: [entity1.id.to_s, entity2.id.to_s]}
          end
          it 'returns status code 200' do
            expect(response).to have_http_status(200)
          end
          it 'returns a valid json' do
            expect(json).not_to be_empty
            expect(json['data'].length).to eq(2)
          end
        end
      end

      ACCEPTABLE_PARAMS.each do |acceptable_param|
        type = MUTATE_PARAMS[acceptable_param.to_sym] || MUTATE_PARAMS[acceptable_param.to_s]
        if type
          describe 'GET /entities with sort on field ' + acceptable_param + ' works ' do
            let!(:entity3) {create(:entity, acceptable_param => get_typed_value(type, 2))}
            let!(:entity1) {create(:entity, acceptable_param => get_typed_value(type, 0))}
            let!(:entity2) {create(:entity, acceptable_param => get_typed_value(type, 1))}
            before :each do
              authenticated_header(request, responsible)
              get 'index', {sort: {acceptable_param => 1}}
            end
            it 'returns sorted results' do
              expect(response.status).to eq(200)
              if type == 'DateTime'
                entity_attribute = json['data'].map {|unit| DateTime.parse(unit['attributes'][acceptable_param])}
                expect(entity_attribute).to eq([get_typed_value(type, 0), get_typed_value(type, 1), get_typed_value(type, 2)])
              elsif type == 'Date'
                entity_attribute = json['data'].map {|unit| Date.parse(unit['attributes'][acceptable_param])}
                expect(entity_attribute).to eq([get_typed_value(type, 0), get_typed_value(type, 1), get_typed_value(type, 2)])
              else
                entity_attribute = json['data'].map {|unit| unit['attributes'][acceptable_param]}
                expect(entity_attribute).to eq([get_typed_value(type, 0), get_typed_value(type, 1), get_typed_value(type, 2)])
              end
            end
          end
        end
      end

      describe 'GET /entities with operators' do
        let!(:entity1) {create(:entity, siret: 'Super', spent_on_completed_requests: 50)}
        let!(:entity3) {create(:entity, siret: 'Super', spent_on_completed_requests: 150)}
        let!(:entity2) {create(:entity, siret: 'bel appart', spent_on_completed_requests: 150)}
        describe 'with $lt' do
          before :each do
            authenticated_header(request, responsible)
            get 'index', {spent_on_completed_requests: {'$lt': 100}}
          end
          it 'returns results' do
            expect(response.status).to eq(200)
            expect(json['meta']['pagination']['total_objects']).to eq(1)
          end
        end
        describe 'with $in' do
          before :each do
            authenticated_header(request, responsible)
            get 'index', {siret: {'$in': ['bel appart', 'flamand']}}
          end
          it 'returns results' do
            expect(response.status).to eq(200)
            expect(json['meta']['pagination']['total_objects']).to eq(1)
          end
        end
        describe 'with $ne' do
          before :each do
            authenticated_header(request, responsible)
            get 'index', {spent_on_completed_requests: {'$ne': 50}}
          end
          it 'returns sorted results' do
            expect(response.status).to eq(200)
            expect(json['meta']['pagination']['total_objects']).to eq(2)
          end
        end
        describe 'with $regex' do
          before :each do
            authenticated_header(request, responsible)
            get 'index', {siret: {'$regex': /^S/}}
          end
          it 'returns sorted results' do
            expect(response.status).to eq(200)
            expect(json['meta']['pagination']['total_objects']).to eq(2)
          end
        end
      end
    end

    describe 'GET /entities/:id' do
      describe 'verif it work' do
        let!(:entity1) {create(:entity)}
        before :each do
          authenticated_header(request, responsible)
          get 'show', {id: entity1.id}
        end
        it 'returns status code 200' do
          expect(response).to have_http_status(200)
        end
        it 'returns a valid json' do
          expect(json).not_to be_empty
        end
      end
      describe 'verif it work not work with a non authorized user' do
        let!(:entity1) {create(:entity)}

        before :each do
          authenticated_header(request, user)
          get 'show', {id: entity1.id}
        end
        it 'returns status code 404' do
          expect(response).to have_http_status(404)
        end
      end
    end

    describe 'PATCH /entities/:id' do
      entity_params = {}
      MUTATE_PARAMS.each do |param, type|
        entity_params[param] = get_typed_value(type)
      end

      let!(:entity) {create(:entity, **entity_params)}
      MUTATE_PARAMS.each do |param, type|
        describe 'patch request with ' + param.to_s + ' ' + type.to_s + 'is valid' do
          let(:params) do
            attr = {}
            attr[param] = get_typed_value(type, 2)
            attr
          end
          before :each do

            authenticated_header(request, responsible)
            patch 'update', {
                id: entity.id.to_s,
                data: {
                    attributes: params
                }
            }
          end
          it 'returns status code 200 and the value appears in json' do
            expect(response).to have_http_status(200)
            if (type == 'Date' || type == 'DateTime')

              expect(json['data']['attributes'][param.to_s]).not_to be_empty
              expect(json['data']['attributes'][param.to_s]).not_to eq('')
            else
              expect(json['data']['attributes'][param.to_s]).to eq(get_typed_value(type, 2))
            end
          end
        end
      end

      describe 'DELETE /entities/:id' do
        let!(:entity) {create(:entity, user_id: responsible.id.to_s)}
        before :each do
          authenticated_header(request, responsible)

          delete 'destroy', {
              id: entity.id.to_s
          }
        end
        it 'returns status code 404' do
          expect(response).to have_http_status(204)
        end
      end
    end
  end

  describe Api::Dmo::EntitiesController, type: :request, swagger_doc: 'api/swagger/hermes/swagger_hermes.json' do
    require_relative '../../swagger_helper'
    TAGS_API_DMO_ENTITIES = 'API_DMO_ENTITIES'
    include JwtAuthHelper
    extend HermesSwaggerJsonParams
    include SwaggerJsonAttributes
    let!(:responsible) {create(:user, first_name: 'Martin', last_name: 'Durand', email: 'martin.durand@gmail.com', roles: ['Administrator'])}
    let!(:user) {create(:user, first_name: 'Jackie', last_name: 'Mike', email: 'jackie.mike@gmail.com')}

    context 'as admin' do
      let!(:admin) {create(:administrator)}

      path '/api/dmo/entities' do
        post 'Create entity' do
          tags TAGS_API_DMO_ENTITIES
          produces 'application/json'
          consumes 'application/json'
          security [Bearer: {}]
          json_params = {}
          MUTATE_PARAMS.each do |key, val|
            json_params[key] = {type: nice_type(val)}
          end
          json_params[:lessor_id] = {type: :string}
          json_params[:housing_offer_ids] = {type: :array}
          json_params[:applicant_ids] = {type: :array}
          parameter name: :params, in: :body, schema: jsonify_params(json_params)
          #jsonapi_params( **json_params )

          response '201', 'entity created' do
            let(:Authorization) {"Bearer #{token_for(responsible)}"}
            let(:params) do
              jsonapi_attributes(
              {}
              )
            end
            include_context 'with integration test'
            include_context 'valid jsonapi'
            include_context 'not containings errors'

            after do |_example|
              json = JSON.parse(response.body)

              expect(json['data']).to be_a Hash
              expect(json['data']['id']).to_not be nil
              expect(json['data']['type']).to eq 'entities'
            end
          end
        end
      end

      path '/api/dmo/entities' do
        get 'Retrieves all entities' do
          tags TAGS_API_DMO_ENTITIES
          produces 'application/json'

          parameter name: 'ids[]', in: :query, type: :array, items: { type: :string }, collectionFormat: :multi, required: false
          ACCEPTABLE_PARAMS.each do |acceptable_param|
            type=''
            if MUTATE_PARAMS.include?(acceptable_param)
              type = nice_type(MUTATE_PARAMS[acceptable_param])
            end
            parameter name: acceptable_param.to_s, in: :query, type: type, required: false
          end
          # VALID_OPERATORS.each do |valid_operator|
          #   parameter( name: valid_operator.to_s, in: :query,type: :object, required: false)
          #end

          security [Bearer: {}]
          paginator

          response '200', 'entities found' do
            let(:Authorization) {"Bearer #{token_for(admin)}"}
            let!(:entities) {create_list(:entity, 5)}

            include_context 'with integration test'
            include_context 'valid jsonapi collection' do
              let(:expected_collection) {entities}
            end
          end
        end
      end

      path '/api/dmo/entities/{id}' do
        get 'Retrieves a single entity by id' do
          tags TAGS_API_DMO_ENTITIES
          produces 'application/json'
          parameter name: :id, in: :path, type: :string
          security [Bearer: {}]

          response '200', 'entity found' do
            let(:Authorization) {"Bearer #{token_for(admin)}"}
            let(:entity) {create(:entity)}
            let(:id) {entity.id}

            include_context 'with integration test'
            include_context 'valid jsonapi ressource' do
              let(:expected_ressource) {entity}
            end
          end
        end
      end

      path '/api/dmo/entities/{id}' do
        put 'edit entity' do
          tags TAGS_API_DMO_ENTITIES
          produces 'application/json'
          consumes 'application/json'
          parameter name: :id, in: :path, type: :string
          security [Bearer: {}]
          json_params = {}
          MUTATE_PARAMS.each do |key, val|
            json_params[key] = {type: nice_type(val)}
          end
          json_params[:lessor_id] = {type: :string}
          json_params[:housing_offer_ids] = {type: :array}
          json_params[:applicant_ids] = {type: :array}
          parameter name: :params, in: :body, schema: jsonify_params(json_params)
          #jsonapi_params( **json_params )

          response '200', 'entity updated' do
            let(:Authorization) {"Bearer #{token_for(admin)}"}
            let(:entity) {create(:entity)}
            let(:id) {entity.id}
            let(:params) do
              jsonapi_attributes(
               {title: 'String',}
              )
            end
            include_context 'with integration test'
            include_context 'valid jsonapi'
            include_context 'not containings errors'

            after do |_example|
              json = JSON.parse(response.body)

              expect(json['data']).to be_a Hash
              expect(json['data']['id']).to_not be nil
              expect(json['data']['type']).to eq 'entities'
            end
          end
        end
      end

      path '/api/dmo/entities/{id}' do
        delete 'Delete a single entity by id' do
          tags TAGS_API_DMO_ENTITIES
          produces 'application/json'
          parameter name: :id, in: :path, type: :string
          security [Bearer: {}]

          response '204', 'entity deleted' do
            let(:Authorization) {"Bearer #{token_for(admin)}"}
            let(:id) {create(:entity, responsible: admin.id).id}
            run_test!
          end
        end
      end
    end
  end
end
